/*
 * Copyright (c) 2008, Marcus von Appen
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <oleacc.h>
#include "papi_msaa.h"
#include "papi_msaaprivate.h"

#define IS_SELF(child) \
    ((((VARIANT) child).vt == VT_I4) && \
     (((VARIANT) child).lVal == CHILDID_SELF))

#define ISACCESSIBLE(x) (PyObject_TypeCheck(x, &PyIAccessible_Type))

static void
_ClientToScreen (HWND hWnd, LPRECT lpRect)
{
    if (lpRect)
    {
        POINT p1 = { lpRect->left, lpRect->top };
        POINT p2 = { lpRect->right, lpRect->bottom };
        ClientToScreen (hWnd, &p1);
        ClientToScreen (hWnd, &p2);
        SetRect (lpRect, p1.x, p1.y, p2.x, p2.y);
    }
}

IAccProxy::IAccProxy (PyObject *self)
{
    this->papiobj = self;
    this->isroot = false;
    this->refcnt = 0;
    this->AddRef ();
}

IAccProxy::~IAccProxy ()
{
    /* Safety net - IUnknown manages the deconstruction. */
    assert (this->refcnt == 0);
}

void
IAccProxy::getAccessible (VARIANT varChild, IAccProxy **child)
{
    *child = NULL;

    if (varChild.vt == VT_DISPATCH)
    {
        IDispatch *iface = varChild.pdispVal;
        /* Evil, but accSelect (and maybe others) receive a wrong(?) vt,
         * which has a VT_DISPATCH spec, but something else as value.
         */
        if (iface != NULL)
            *child = static_cast<IAccProxy*>(iface);
        else
            varChild.vt = VT_I4; 
    }
    
    if (varChild.vt == VT_I4)
    {
        IDispatch *iface;

        if (varChild.lVal == CHILDID_SELF)
        {
            *child = this;
            return;
        }

        this->get_accChild (varChild, &iface);
        if (iface != NULL)
            *child = static_cast<IAccProxy*>(iface);
    }
    return;
}

void
IAccProxy::ResetPythonRef ()
{
    this->papiobj = NULL;
}

/* IAccessible interfaces */

STDMETHODIMP
IAccProxy::accDoDefaultAction (VARIANT varChild)
{
    IAccProxy *p = NULL;
    debug ("accDoDefaultAction\n");

    this->getAccessible (varChild, &p);
    if (p)
    {
        PyObject *retval = PyObject_CallMethod (p->papiobj,
            "accDoDefaultAction", NULL);
        if (!retval)
            return DISP_E_MEMBERNOTFOUND;

        Py_DECREF (retval);
        return S_OK;
    }
    return E_FAIL;
}

STDMETHODIMP
IAccProxy::accHitTest (long xLeft, long yTop, VARIANT *pvarChild)
{
    PyObject *retval;
    POINT px;

    debug ("accHitTest\n");

    if (!pvarChild)
        return E_INVALIDARG;

    pvarChild->vt = VT_EMPTY;

    px.x = xLeft;
    px.y = yTop;
    if (!ScreenToClient (this->hwnd, &px))
        return E_FAIL;

    retval = PyObject_CallMethod (this->papiobj, "accHitTest", "ii",
        px.x, px.y);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (retval == this->papiobj)
    {
        /* It's not within a child */
        pvarChild->vt = VT_I4;
        pvarChild->lVal = CHILDID_SELF;
        this->AddRef ();
    }
    else if (ISACCESSIBLE (retval))
    {
        pvarChild->vt = VT_DISPATCH;
        pvarChild->pdispVal = static_cast<IDispatch*>(IACCPROXY (retval));
        IACCPROXY (retval)->AddRef ();
    }
    else
    {
        Py_DECREF (retval);
        return E_FAIL;
    }
    Py_DECREF (retval);
    return S_OK;
}

STDMETHODIMP
IAccProxy::accLocation (long *pxLeft, long *pyTop, long *pcxWidth,
    long *pcyHeight, VARIANT varChild)
{
    RECT rect;
    PyObject *retval = NULL;
    IAccProxy *p = NULL;

    debug ("accLocation\n");

    if (!pxLeft || !pyTop || !pcxWidth || !pcyHeight)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "accLocation", NULL);
    if (!retval || retval == Py_None)
    {
        Py_XDECREF (retval);
        if (GetClientRect (p->hwnd, &rect))
        {
            _ClientToScreen (p->hwnd, &rect);
            *pxLeft = rect.left;
            *pyTop = rect.top;
            *pcxWidth = rect.right - rect.left;
            *pcyHeight = rect.bottom - rect.top;
            return S_OK;
        }
        return S_FALSE;
    }

    if (PySequence_Check (retval) && PySequence_Size (retval) == 4)
    {
        rect.left = PyInt_AsLong (PySequence_Fast_GET_ITEM (retval, 0));
        if (rect.left == -1 && PyErr_Occurred ())
        {
            PyErr_Clear ();
            Py_DECREF (retval);
            return E_FAIL;
        }

        rect.top = PyInt_AsLong (PySequence_Fast_GET_ITEM (retval, 1));
        if (rect.top == -1 && PyErr_Occurred ())
        {
            PyErr_Clear ();
            Py_DECREF (retval);
            return E_FAIL;
        }

        rect.right = PyInt_AsLong (PySequence_Fast_GET_ITEM (retval, 2));
        if (rect.right== -1 && PyErr_Occurred ())
        {
            PyErr_Clear ();
            Py_DECREF (retval);
            return E_FAIL;
        }
        else
            rect.right += rect.left;

        rect.bottom = PyInt_AsLong (PySequence_Fast_GET_ITEM (retval, 3));
        if (rect.bottom == -1 && PyErr_Occurred ())
        {
            PyErr_Clear ();
            Py_DECREF (retval);
            return E_FAIL;
        }
        else
            rect.bottom += rect.top;

        _ClientToScreen (p->hwnd, &rect);
        *pxLeft = rect.left;
        *pyTop = rect.top;
        *pcxWidth = rect.right - rect.left;
        *pcyHeight = rect.bottom - rect.top;
    }
    Py_DECREF (retval);
    return S_OK;
}

STDMETHODIMP
IAccProxy::accNavigate (long navDir, VARIANT varStart, VARIANT *pvarEndUpAt)
{
    PyObject *retval;
    IAccProxy *p = NULL;

    debug ("accNavigate\n");

    if (!pvarEndUpAt)
        return E_INVALIDARG;

    this->getAccessible (varStart, &p);
    if (!p)
    {
        pvarEndUpAt->vt = VT_EMPTY;
        return S_FALSE;
    }

    retval = PyObject_CallMethod (p->papiobj, "accNavigate", "i", navDir);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (retval == Py_None)
    {
        Py_DECREF (retval);
        pvarEndUpAt->vt = VT_EMPTY;
        return S_FALSE;
    }
    else if (ISACCESSIBLE (retval))
    {
        pvarEndUpAt->vt = VT_DISPATCH;
        pvarEndUpAt->pdispVal = static_cast<IDispatch*>(IACCPROXY (retval));
        IACCPROXY (retval)->AddRef ();
        Py_DECREF (retval);
        return S_OK;
    }
    return S_FALSE;
}

STDMETHODIMP
IAccProxy::accSelect (long flagsSelect, VARIANT varChild)
{
    HRESULT result = S_FALSE;
    PyObject *retval;
    IAccProxy *p = NULL;
    debug ("accSelect\n");

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "accSelect", "l", flagsSelect);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (retval == Py_True)
        result = S_OK;

    Py_DECREF (retval);
    return result;
}

STDMETHODIMP
IAccProxy::get_accChild (VARIANT varChild, IDispatch **ppdispChild)
{
    PyObject *retval;

    debug ("get_accChild\n");

    if (!ppdispChild)
        return E_INVALIDARG;

    if (varChild.vt != VT_I4)
        return E_FAIL;

    if (varChild.lVal == CHILDID_SELF)
    {
        *ppdispChild = static_cast<IDispatch*>(this);
        this->AddRef ();
        return S_OK;
    }

    retval = PyObject_CallMethod (this->papiobj, "get_accChild", "i",
        varChild.lVal);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (ISACCESSIBLE (retval))
    {
        *ppdispChild = static_cast<IDispatch*>(IACCPROXY (retval));
        IACCPROXY (retval)->AddRef ();
        Py_DECREF (retval);
        return S_OK;
    }

    *ppdispChild = NULL;
    Py_DECREF (retval);
    return S_FALSE;
}

STDMETHODIMP
IAccProxy::get_accChildCount (long *pcountChildren)
{
    PyObject *retval = NULL;

    debug ("get_accChildCount\n");

    if (!pcountChildren)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (this->papiobj, "get_accChildCount", NULL);
    if (!retval) /* Fine, if pcountChildren == NULL */
        return DISP_E_MEMBERNOTFOUND;

    *pcountChildren = PyInt_AsLong (retval);
    Py_DECREF (retval);
    if (*pcountChildren == -1 && PyErr_Occurred ())
    {
        PyErr_Clear ();
        *pcountChildren = 0;
        return S_FALSE;
    }
    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accDefaultAction (VARIANT varChild, BSTR *pszDefaultAction)
{
    char *val;
    PyObject *retval;
    OLECHAR *sOleText;
    IAccProxy *p = NULL;

    debug ("get_accDefaultAction\n");
    if (!pszDefaultAction)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accDefaultAction", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return S_FALSE;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszDefaultAction = ::SysAllocString (sOleText);
    delete sOleText;
    Py_DECREF (retval);

    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accDescription (VARIANT varChild, BSTR *pszDescription)
{
    char *val;
    PyObject *retval;
    OLECHAR *sOleText;
    IAccProxy *p = NULL;

    debug ("get_accDescription\n");
    if (!pszDescription)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accDescription", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return S_FALSE;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszDescription = ::SysAllocString (sOleText);
    delete sOleText;
    Py_DECREF (retval);

    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accFocus (VARIANT *pvarChild)
{
    HWND focused;
    HRESULT hr;

    debug ("get_accFocus\n");
    if(!pvarChild)
        return E_INVALIDARG;

    focused = GetFocus ();
    if (this->hwnd == focused)
    {
        PyObject *retval = PyObject_CallMethod (this->papiobj, "get_accFocus",
            NULL);
        if (!retval)
            return DISP_E_MEMBERNOTFOUND;

        if (ISACCESSIBLE (retval))
        {
            if (retval == this->papiobj)
            {
                pvarChild->vt = VT_I4;
                pvarChild->lVal = CHILDID_SELF;
                this->AddRef ();
            }
            else
            {
                pvarChild->vt = VT_DISPATCH;
                pvarChild->pdispVal =
                    static_cast<IDispatch*>(IACCPROXY (retval));
                IACCPROXY (retval)->AddRef ();
            }
            hr = S_OK;
        }
        else
        {
            pvarChild->vt = VT_EMPTY;
            hr = S_FALSE;
        }
        Py_DECREF (retval);
    }
    else
    {
        pvarChild->vt = VT_DISPATCH;
        hr = AccessibleObjectFromWindow (focused, OBJID_WINDOW,
            IID_IAccessible, (void**) &(pvarChild->pdispVal));
        hr = S_FALSE;
    }
    return hr;
}

STDMETHODIMP
IAccProxy::get_accHelp (VARIANT varChild, BSTR *pszHelp)
{
    char *val;
    PyObject *retval;
    OLECHAR *sOleText;
    IAccProxy *p = NULL;
    debug ("get_accHelp\n");

    if (!pszHelp)
        return E_INVALIDARG;
    *pszHelp = NULL;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accHelp", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (retval == Py_None)
    {
        Py_DECREF (retval);
        return S_FALSE;
    }

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return S_FALSE;
    }
    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszHelp = ::SysAllocString (sOleText);
    delete sOleText;
    Py_DECREF (retval);
    if (!(*pszHelp))
        return S_FALSE;

    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accHelpTopic (BSTR *pszHelpFile, VARIANT varChild,long *pidTopic)
{
    char *val;
    PyObject *obj;
    PyObject *retval;
    OLECHAR *sOleText;
    IAccProxy *p = NULL;

    debug ("get_accHelp\n");

    if (!pszHelpFile || !pidTopic)
        return E_INVALIDARG;

    *pszHelpFile = NULL;
    *pidTopic = 0;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accHelpTopic", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (!PySequence_Check (retval) || PySequence_Size (retval) != 2)
    {
        Py_DECREF (retval);
        return S_FALSE;
    }

    obj = PySequence_GetItem (retval, 1);
    *pidTopic = PyLong_AsLong (obj);
    Py_DECREF (obj);
    if (*pidTopic == -1 && PyErr_Occurred ())
    {
        PyErr_Clear ();
        *pidTopic = 0;
        Py_DECREF (retval);
        return E_FAIL;
    }
    Py_DECREF (retval);

    obj = PySequence_GetItem (retval, 0);
    val = PyString_AsString (obj);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        Py_DECREF (obj);
        return E_FAIL;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszHelpFile = ::SysAllocString (sOleText);
    delete sOleText;
    Py_DECREF (obj);
    if (!(*pszHelpFile))
        return S_FALSE;

    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accKeyboardShortcut (VARIANT varChild, BSTR *pszKeyboardShortcut)
{
    IAccProxy *p = NULL;
    char *val;
    PyObject *retval;
    OLECHAR *sOleText;

    if (!pszKeyboardShortcut)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accKeyboardShortcut", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return S_FALSE;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszKeyboardShortcut = ::SysAllocString (sOleText);
    Py_DECREF (retval);
    delete sOleText;
    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accName (VARIANT varChild, BSTR *pszName)
{
    IAccProxy *p = NULL;
    char *val;
    PyObject *retval;
    OLECHAR *sOleText;

    debug ("get_accName\n");
    if (!pszName)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval =  PyObject_CallMethod (p->papiobj, "get_accName", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    if (retval == Py_None)
    {
        Py_DECREF (retval);
        return S_FALSE;
    }

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return S_FALSE;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszName = ::SysAllocString (sOleText);
    Py_DECREF (retval);
    delete sOleText;
    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accParent (IDispatch **ppdispParent)
{
    HRESULT hr = S_FALSE;
    PyObject *retval;

    debug ("get_accParent\n");

    if (!ppdispParent)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (this->papiobj, "get_accParent", NULL);
    if (!retval || retval == Py_None)
    {
        if (this->isroot)
        {
            AccessibleObjectFromWindow(this->hwnd, (DWORD) OBJID_WINDOW,
                IID_IAccessible, (void**) ppdispParent);
            hr = (*ppdispParent) ? S_OK : S_FALSE;
        }
        else
        {
            HWND handle = GetParent (this->hwnd);
            if (!handle)
                handle = GetDesktopWindow ();
            AccessibleObjectFromWindow(handle, (DWORD) OBJID_CLIENT,
                IID_IAccessible, (void**) ppdispParent);
            hr = (*ppdispParent) ? S_OK : S_FALSE;
        }
    }
    else if (ISACCESSIBLE (retval))
    {
        *ppdispParent = static_cast<IDispatch*>(IACCPROXY (retval));
        IACCPROXY (retval)->AddRef ();
        hr = S_OK;
    }
    Py_XDECREF (retval);
    return hr;
}

STDMETHODIMP
IAccProxy::get_accSelection (VARIANT *pvarChildren)
{
    PyObject *retval;

    debug ("get_accSelection\n");
    if (!pvarChildren)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (this->papiobj, "get_accSelection", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    /* Single accessible objet? */
    if (ISACCESSIBLE (retval))
    {
        if (IACCPROXY (retval) == this)
        {
            pvarChildren->vt = VT_I4;
            pvarChildren->lVal = CHILDID_SELF;
            this->AddRef ();
        }
        else
        {
            pvarChildren->vt = VT_DISPATCH;
            pvarChildren->pdispVal =
                static_cast<IDispatch*>(IACCPROXY (retval));
            IACCPROXY (retval)->AddRef ();
        }
    }
    else if (PySequence_Check (retval))
    {
        /* TODO */
        pvarChildren->vt = VT_EMPTY;
        return S_FALSE;
    }

    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accState (VARIANT varChild, VARIANT *pvarState)
{
    PyObject *retval;
    IAccProxy *p = NULL;
    long sval = 0;
    debug ("get_accState\n");

    if (!pvarState)
        return E_INVALIDARG;
    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    pvarState->vt = VT_I4;
    retval = PyObject_CallMethod (p->papiobj, "get_accState", NULL);
    if (!retval)
    {
        pvarState->lVal = 0;
        return E_FAIL;
    }
    sval = PyLong_AsLong (retval);
    Py_DECREF (retval);
    if (sval == -1 && PyErr_Occurred ())
    {
        PyErr_Clear ();
        pvarState->lVal = 0;
        return E_FAIL;
    }
    pvarState->lVal = sval;
    return S_OK;
}

STDMETHODIMP
IAccProxy::get_accValue (VARIANT varChild, BSTR *pszValue)
{
    char *val;
    PyObject *retval;
    IAccProxy *p = NULL;
    OLECHAR *sOleText;
    debug ("get_accValue\n");

    if (!pszValue)
        return E_INVALIDARG;
    *pszValue = NULL;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accValue", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;
    if (retval == Py_None)
    {
        Py_DECREF (retval);
        return S_FALSE;
    }

    val = PyString_AsString (retval);
    if (!val)
    {
        PyErr_Clear ();
        Py_DECREF (retval);
        return E_FAIL;
    }

    sOleText = new OLECHAR[strlen (val) + 1];
    mbstowcs (sOleText, val, strlen (val) + 1);
    *pszValue = ::SysAllocString (sOleText);
    Py_DECREF (retval);
    delete sOleText;

    return S_OK;
}

STDMETHODIMP
IAccProxy::put_accName (VARIANT varChild, BSTR szName)
{
    return E_NOTIMPL;
}

STDMETHODIMP
IAccProxy::put_accValue (VARIANT varChild, BSTR szValue)
{
    return E_NOTIMPL;
}

STDMETHODIMP
IAccProxy::get_accRole (VARIANT varChild, VARIANT *pvarRole)
{
    PyObject *retval;
    IAccProxy *p = NULL;

    debug ("get_accRole\n");
    if (!pvarRole)
        return E_INVALIDARG;

    this->getAccessible (varChild, &p);
    if (!p)
        return E_INVALIDARG;

    retval = PyObject_CallMethod (p->papiobj, "get_accRole", NULL);
    if (!retval)
        return DISP_E_MEMBERNOTFOUND;

    pvarRole->vt = VT_I4;
    pvarRole->lVal = PyInt_AsLong (retval);
    Py_DECREF (retval);
    if (pvarRole->lVal == -1 && PyErr_Occurred ())
    {
        PyErr_Clear ();
        pvarRole->vt = VT_EMPTY;
        return S_FALSE;
    }
    return S_OK;
}

/* IUnknown interface implementations */

STDMETHODIMP
IAccProxy::QueryInterface (REFIID riid, void** ppvObject)
{
    debug ("QueryInterface\n");
    if (riid == IID_IAccessible)
        *ppvObject = static_cast<IAccessible*>(this);
    else if (riid == IID_IDispatch)
        *ppvObject = static_cast<IDispatch*>(this);
    else if (riid == IID_IUnknown)
        *ppvObject = static_cast<IUnknown*>(this);
    else
        *ppvObject = NULL;

    if (*ppvObject)
    {
        reinterpret_cast<IUnknown*>(*ppvObject)->AddRef ();
        return S_OK;
    }
    return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG)
IAccProxy::AddRef ()
{
    debug ("AddRef\n");
    return InterlockedIncrement ((LONG*) &(this->refcnt));
}

STDMETHODIMP_(ULONG)
IAccProxy::Release ()
{
    ULONG ref;

    debug ("Release\n");
    ref = InterlockedDecrement ((LONG*) &(this->refcnt));

    if (ref == 0)
    {
        debug ("Release refcount == 0\n");
        if (this->papiobj)
            Py_DECREF (this->papiobj);
        delete this;
    }
    return this->refcnt;
}

/* IDispatch interfaces
 *
 * We do not need to take care about it as MSAA 2.0 takes care of it
 * according to the MSDN.
 */
STDMETHODIMP
IAccProxy::GetTypeInfoCount (unsigned int FAR* pctinfo)
{
    debug ("GetTypeInfoCount\n");
    return E_NOTIMPL;
}

STDMETHODIMP
IAccProxy::GetTypeInfo (unsigned int iTInfo, LCID lcid,
    ITypeInfo FAR* FAR* ppTInfo)
{
    debug ("GetTypeInfo\n");
    return E_NOTIMPL;
}

STDMETHODIMP
IAccProxy::GetIDsOfNames (REFIID riid, OLECHAR FAR* FAR* rgszNames,
    unsigned int cNames, LCID lcid, DISPID FAR* rgDispId)
{
    debug ("GetIDsOfNames\n");
    return E_NOTIMPL;
}

STDMETHODIMP
IAccProxy::Invoke (DISPID dispIdMember, REFIID riid, LCID lcid, WORD wFlags,
    DISPPARAMS FAR* pDispParams, VARIANT FAR* pVarResult,
    EXCEPINFO FAR* pExcepInfo, unsigned int FAR* puArgErr)
{
    debug ("Invoke\n");
    return E_NOTIMPL;
}
