/*
 * Copyright (c) 2008, Marcus von Appen
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "papi_msaa.h"
#include "papi_msaaprivate.h"
#include "papi_iaccproxy.h"

#include <tchar.h>
#include <windows.h>
#include <windowsx.h>

#ifdef STRICT
#define WNDPROCTYPE WNDPROC
#else
#define WNDPROCTYPE FARPROC
#endif

/**
 * Getter method used as wrapper for get_root() and the hook for the
 * window messaging interface. It will be bound by user code usually
 * using something like:
 *
 * papi.get_root = my_main_widget.get_accessible
 *
 */
static PyObject *_get_root = NULL;

/**
 * Original WNDPROC message loop for the _msaa_hook_on method. This
 * preserves the window's original message loop and lets us forward
 * messages to it, that are not handled by our own WNDPROC.
 */
static WNDPROCTYPE _wndproc = NULL;

/**
 * The window handle the MSAA implementation belongs to.
 */
static HWND _hwnd = 0;

/* Forward declarations. */
static LRESULT CALLBACK _msaa_wndproc (HWND hWnd, UINT nMsg,WPARAM wParam,
    LPARAM lParam);
static PyObject* _msaa_hook_on (PyObject *self, PyObject *args);
static PyObject* _msaa_set_root (PyObject *self, PyObject *args);
static PyObject* _msaa_get_root (PyObject *self);
static IAccProxy* _iacc_get_root (void);

/**
 * WNDPROC implementation that explicitly handles the WM_GETOBJECT
 * message to return the set root accessible.
 */
static LRESULT CALLBACK
_msaa_wndproc (HWND hWnd, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
    if (nMsg == WM_GETOBJECT && lParam == OBJID_CLIENT)
    {
        IAccProxy *proxy = _iacc_get_root ();
        if (proxy)
        {
            proxy->isroot = true;
            return LresultFromObject (IID_IAccessible, wParam,
                static_cast<IAccessible*>(proxy));
        }
        return 0;
    }
    if (_wndproc)
        return CallWindowProc (_wndproc, hWnd, nMsg, wParam, lParam);
    return DefWindowProc (hWnd, nMsg, wParam, lParam);
}

/**
 * papi.msaa.hook_on implementation.
 * Exchanges the passed window's WNDPROC with our own one.
 */
static PyObject*
_msaa_hook_on (PyObject *self, PyObject *args)
{
    long windowid;

    if (!PyArg_ParseTuple (args, "l:hook_on", &windowid))
        return NULL;

    _wndproc = (WNDPROCTYPE) GetWindowLongPtr ((HWND) windowid, GWLP_WNDPROC);
    if (_wndproc == NULL)
        Py_RETURN_FALSE;

    SetWindowLongPtr ((HWND) windowid, GWLP_WNDPROC,
        (LONG_PTR) _msaa_wndproc);
    _hwnd = (HWND) windowid;

    Py_RETURN_TRUE;
}

/**
 * papi.msaa.get_root implementation.
 */
static PyObject*
_msaa_set_root (PyObject *self, PyObject *args)
{
    PyObject *tmp;

    debug ("_msaa_set_root\n");

    if (!PyArg_ParseTuple (args, "O:set_root", &tmp))
        return NULL;

    if (!PyCallable_Check (tmp)
        && !PyObject_TypeCheck (tmp, &PyIAccessible_Type))
    {
        PyErr_SetString (PyExc_TypeError,
            "parameter must be callable or an IAccessible");
        return NULL;
    }
    Py_XINCREF (tmp);
    Py_XDECREF (_get_root);
    _get_root = tmp;
    Py_RETURN_NONE;
}

/**
 * Python getter for the root object.
 */
static PyObject*
_msaa_get_root (PyObject *self)
{
    PyObject *result;

    debug ("_msaa_get_root\n");

    if (!_get_root)
    {
        PyErr_SetString (PyExc_NameError, "Interface get_root not satisfied.");
        return NULL;
    }

    if (!PyCallable_Check (_get_root))
    {
        Py_INCREF (_get_root);
        return _get_root;
    }

    result = (PyObject *) PyEval_CallObject (_get_root, NULL);
    if (!result)
        return NULL;

    Py_INCREF (result);
    return result;
}

/**
 * Gets the IAccessible from the bound Python root method or object.
 */
static IAccProxy*
_iacc_get_root (void)
{
    PyObject *result;
    IAccProxy *p = NULL;

    debug ("_iacc_get_root\n");

    if (!_get_root)
    {
        return NULL;
    }
    if (!PyCallable_Check (_get_root))
    {
        p = IACCPROXY (_get_root);
        if (!p)
            return NULL;
        p->hwnd = _hwnd;
        return p;
    }
    result = (PyObject *) PyEval_CallObject (_get_root, NULL);
    if (!result)
        return NULL;
    Py_DECREF (result);

    p = IACCPROXY (result);
#ifdef DEBUG
    if (!p)
        debug ("_iacc_get_root has no IACCPROXY\n");
#endif
    p->hwnd = _hwnd;
    return p;
}

HWND
get_root_hwnd (void)
{
    return _hwnd;
}

/**
 * Python module initialization function. This will be called upon the
 * import <module> statement and sets up the papi module.
 */
PyMODINIT_FUNC
initmsaa (void)
{
    PyObject *mod;
    PyObject *func;
    PyObject *c_api_obj;
    static void *c_api[PAPI_MSAAAPI_SLOTS];
    int i;

    static PyMethodDef methods[] = {
        { "hook_on", _msaa_hook_on, METH_VARARGS,
          "hook_on (windowid) -> bool\n\n"
          "Sets the window to hook the message processing on.\n\n"
          "Return True, if the message processing could be successfully\n"
          "applied to the window, False otherwise." },
        { "get_root", (PyCFunction) _msaa_get_root, METH_NOARGS,
          "get_root -> IAccessible\n\n"
          "Gets the root accessible container for the current application." },
        { "set_root", _msaa_set_root, METH_VARARGS,
          "set_root (callback or object -> None\n\n"
          "Sets the method, that retrieves the root object." },
        { NULL, NULL, 0, NULL }
    };

    /* Complete types. */
    if (PyType_Ready (&PyIAccessible_Type) < 0)
        return;

    mod = Py_InitModule3 ("msaa", NULL, "MSAA module for Python objects.");

    Py_INCREF (&PyIAccessible_Type);
    PyModule_AddObject (mod, "IAccessible", (PyObject *) &PyIAccessible_Type);

    /* Export the MSAA constants. */
    constants_export (mod);

    /* Export the methods. */
    i = 0;
    while (methods[i].ml_name != NULL)
    {
        func = PyCFunction_New (&methods[i], NULL);
        PyObject_SetAttrString (mod, methods[i].ml_name, func);
        i++;
    }

    /* C API exorts */
    iaccessible_export_api (c_api);
    c_api_obj = PyCObject_FromVoidPtr ((void *) c_api, NULL);
    if (c_api_obj)
        PyModule_AddObject (mod, PAPI_MSAACAPI_ENTRY, c_api_obj);

    if (CoInitialize (NULL) != S_OK)
        PyErr_SetString (PyExc_WindowsError,
            "COM system could not be initialized\n");
}
